import os
import logging
from uuid import uuid1

from croniter import croniter
from datetime import datetime, date

from p2.driver.k8s.k8s import Cron
from p2.driver.k8s.jobs import Job
from p2.store.client import Client as StoreClient
from p2.scheduler.Rule import RuleEngine
from p2.scheduler.Config import Config
from p2.scheduler.Test import P2test
from p2.scheduler.Shifts import Shifts
from p2.logging import HttpLogger
from p2.tools.packages import get_p2_packages


class Scheduler:
    """A class to schedule testing campaigns.
    The workflow is roughly divided into these steps:
    - First we get a list of what tests (experiments on targets) we would like to run
    - Then we sort all tests executing the set of rules defined for this scheduler.
    - Then we try to find a free slot to run them
    - Lastly we schedule the execution of these tests as Kubernetes cronjobs in the same
      context the scheduler is running in.

    A scheduler object accepts as attributes:
    - a tag: (not mandatory) used to disambiguate between multiple scheduler instances.
    """

    def __init__(self, tag: str):

        try:
            self.store_url = os.environ.get('P2_STORE_URL')
            self.p2_store_client = StoreClient(self.store_url)
            if self.p2_store_client is None:
                raise Exception(f"Error getting a p2.store client object [{self.store_url}]")

            log_obj = HttpLogger(
                level   = logging.DEBUG,
                silent  = False,
                name    = f"p2.scheduler.{tag}",
                p2_store_url = self.store_url
            )
            self.log = log_obj.getLogger()

            self.context = tag
            self.config = Config(tag = tag)
            self.shifts = Shifts(tag = tag)
            self.rule_engine = RuleEngine(self.config, self.log)
            self.to_be_scheduled = list()
            self.cron = Cron(context=tag,
                             namespace=self.config.cronjobs.get('namespace'))
            self.cronjobs = [i['name'] for i in self.cron.get_cronjobs()]
            self.job = Job(context=tag,
                           logger = self.log,
                           namespace=self.config.cronjobs.get('namespace'))
            self.jobs = [i['_metadata'].__dict__['_name'] for i in self.job.get_jobs()]
            self.log.info(f"Scheduler object with tag {tag} initialised")
            self.log.info("Installed P2 packages:")
            self.log.info("\n".join(get_p2_packages()))

        except Exception as err:
            self.log.error(err)
            return None


    def schedule_campaign(self, campaign_id: int):
        """
        - retrieves the campaign with id
        - get targets
        - get experiments
        - see if experiments have to be scheduled...
        """
        self.log.info(f"Scheduling experiments for testing campaign: {campaign_id}")
        self.campaign_id = campaign_id
        self.campaign    = self._get_campaign()
        self.targets     = self.p2_store_client.select(
                               table='resources',
                               filter=self.campaign['resources']
                           )
        self.experiments = self.p2_store_client.select(
                               table='experiments',
                               filter=self.campaign['experiments']
                           )

        self._cleanup()
        self._get_schedule()
        self.rule_engine.apply_rules(self.to_be_scheduled)
        self._schedule_all()
        return 1


    def _cleanup(self):
        """
        delete old sessions that passed the expiration date
        as Kubernetes can't store the same amount of cronjobs, cronjobs
        life time is much shorter...
        """
        old_sessions = self.p2_store_client.select(
            table='exp_session',
            filter={'start_time|< now() - interval ': f"{self.config.delete_sessions_after} days"},
            fetch=['exp_id', 'start_time']
        )
        for session in old_sessions:
            try:
                self.p2_store_client.delete(table='exp_session', key='exp_id', value=session['exp_id'])
                self.log.debug(f"Session {session['exp_id']} executed on the {session['start_time']} deleted...")
            except Exception as e:
                self.log.error(f"ERROR deleting session {session['exp_id']}: {e}")

        old_cronjobs = self.p2_store_client.select(
            table='exp_session',
            filter={'start_time|< now() - interval ': f"{self.config.delete_jobs_after} days"},
            fetch=['exp_id', 'start_time']
        )
        if hasattr(self, 'cronjobs'):
            for cronjob in old_cronjobs:
                if cronjob['exp_id'] in self.cronjobs:
                    try:
                        if hasattr(self, 'cron'):
                            self.cron.delete_cronjob(cronjob['exp_id'])
                            self.log.debug(f"""Cronjob {cronjob['exp_id']} executed on the
                                            {cronjob['start_time']} deleted...""")
                    except Exception as e:
                        self.log.error(f"ERROR deleting cronjob {job['exp_id']} on context {self.context}: {e}")

        try:
            self.job.delete_jobs_older_than(self.config.delete_jobs_after)
        except Exception as e:
            self.log.error(f"ERROR deleting old jobs on context {self.context}: {e}")

    def _get_schedule(self):
        for target in self.targets:
            for experiment in self.experiments:
                if  experiment['type'] == 'experiment' and \
                    'profile' in target and \
                    'target' in experiment['profile'] and \
                    'name' in experiment['profile']['target'] and \
                    'name' in target['profile'] and\
                    target['profile']['name'] == experiment['profile']['target']['name']:
                    self._add_to_schedule(target=target, experiment=experiment)
                elif experiment['type'] == 'template':
                    # if it's a template then any target is valid...
                    self._add_to_schedule( target=target, experiment=experiment)


    def _add_to_schedule(self, target: dict, experiment: dict):
        self.log.info("Adding experiment {} on target {} to schedule...".format(
            experiment['filename'], target['name']))
        self.to_be_scheduled.append( P2test(
            config = self.config,
            target = target,
            experiment = experiment
            ))


    def _schedule_all(self):
        """schedule tests sorted by expiretion time"""
        not_scheduled=list()
        to_be_scheduled_sorted = sorted(self.to_be_scheduled,
                                        key=lambda item: item.rank,
                                        reverse=True)
        for test in to_be_scheduled_sorted:
            duration = test.experiment['duration'] if 'duration' in test.experiment else 30
            if (start_at := self.shifts.get_session_shift(duration)) is not None:
                day = date(
                    day=start_at['day'],
                    month=start_at['month'],
                    year=start_at['year']
                )
                self._schedule_this(test, day, start_at)
            else:
                self.log.warning("No days left to schedule experiment {} on target {}".format(
                    test.experiment['filename'], test.target['name']))
                not_scheduled.append(test)
        return not_scheduled


    def _schedule_this(self, test: P2test, day: date, start_at: dict):
        """Finds the best slot in the day for this test

            Gets the max(next_run_time) of the day
            adds that test expected duration and round up to next hour
            check hour is in test shift
            in case of a template:
            - template profile should hold the template static parameters
            in case these are set in the test campaign, they override the ones
            set in the template. This is handled by the agent (cron) job.
            - session profile should hold the dynamic params
            - Current dynamic params supported:
                - target: target profile
            """
        try:
            mins = "%02d" % start_at['start_minute']
            hour = "%02d" % start_at['start_hour']
            dom  = "%02d" % start_at['day']
            month= "%02d" % start_at['month']
            schedule = f"{mins} {hour} {dom} {month} *"
            cron = croniter(schedule, datetime.now())
            next_run_time = cron.get_next(datetime)
            exp_id = test.target['name'] + '-' + str(uuid1())[:8]
            session = {
                    'filepath': test.experiment['filename'],
                    'campaign_id': self.campaign_id,
                    'context': self.context,
                    'schedule': schedule,
                    'next_run_time': str(next_run_time),
                    'mins': mins,
                    'hour': hour,
                    'dom': dom,
                    'month': month,
                    'dow': '*',
                    'year': start_at['year'],
                    'profile':{'target':test.target}, # TODO : this is going to be deleted soon...
                    'resource_id': test.target['id'],
                    'status': 'loaded'
            }
            exp_id = self.p2_store_client.create_session(exp_id, session)
            if exp_id != None:
                command_list = self.create_command(
                    exp_id,
                    self.config.cronjobs['command'],
                    test
                )
                name = self.cron.create_cronjob(
                    filename = test.experiment['filename'],
                    name = exp_id,
                    schedule = schedule,
                    image = self.config.cronjobs['image'],
                    command = command_list,
                    service_account = self.config.cronjobs['service_account']
                )
                if name is None:
                    raise Exception("Couldn't add cronjob")
                else:
                    self.log.info(f"Cronjob {name} created...")

                self.p2_store_client.update_session(exp_id, {'status':'scheduled'})

                if test in self.to_be_scheduled:
                    self.to_be_scheduled.remove(test)

                self.log.info(f"Session {exp_id} scheduled")
            else:
                self.log.error(f"Error scheduling session for experiment {session['filepath']}")
            return exp_id
        except Exception as e:
            self.log.error(f"Couldn't schedule test, exception: {e}")
            return None


    def _get_campaign(self):
        campaign = self.p2_store_client.get_campaign_of_id(self.campaign_id)
        if campaign is None:
            raise Exception(f"No campaign with id {id} found")

        if campaign['resources'] is None:
            raise Exception(f"Error, no resources set for campaign {id}")

        if campaign['experiments'] is None:
            raise Exception(f"Error, no experiments set for campaign {id}")
        return campaign


    def create_command(self, exp_id: str, command: str, test: P2test):
        cmd_array = [
            command,
            '--p2_store_url',
            self.store_url,
            '--filename',
            test.experiment['filename'],
            '--driver',
            test.experiment['type'],
            '--session_id',
            exp_id
        ]
        if self.campaign_id:
            cmd_array.extend([
                '--campaign_id',
                str(self.campaign_id)
            ])
        return cmd_array
