import os
from datetime import datetime, timedelta, date

from p2.scheduler.Config import Config
from p2.scheduler.Calendar import Calendar
from p2.store.client import Client as StoreClient

class Shifts:
    """A class to handle all shifts in the scheduling period.

    It accepts a tag that can be used to disambiguate among same day shifts.
    It exposes a method to find the best start time for a session in a given day
    given its duration"""

    def __init__(self, tag: str = None):

        self.tag = tag
        self.config = Config()
        self.cal = Calendar()
        self.p2_store_client = StoreClient(os.environ.get('P2_STORE_URL'))
        if self.p2_store_client is None:
            raise Exception("Error getting a p2.store client object")

        # initializing the shift structure...
        today = date.today()
        shifts = dict()

        for delta_day in range(self.config.look_ahead_period):
            day = today + timedelta(days=delta_day+1)
            if self.cal._day_valid(day) is False:
                continue
            shifts[day] = self.p2_store_client.select(
                table='shifts',
                filter={
                    'day': day.day,
                    'month': day.month,
                    'year': day.year,
                    'tag': tag
                })

            if shifts[day] is None or len(shifts[day]) == 0:
                # initialize a shift record as per config and store it...
                day_shifts = list()
                shift_id = 1
                for config_shift in self.config.shifts:
                    shift = dict()
                    shift['day'] = day.day
                    shift['month'] = day.month
                    shift['year'] = day.year
                    shift['shift_id'] = shift_id
                    shift_id += 1
                    shift['start_hour'] = config_shift['start']['hour']
                    shift['start_minute'] = config_shift['start']['minute']
                    shift['end_hour'] = config_shift['end']['hour']
                    shift['end_minute'] = config_shift['end']['minute']
                    shift['num_sessions_in_shift'] = 0
                    shift['tag'] = self.tag

                    self.p2_store_client.add(
                        table='shifts',
                        record=shift
                    )
                    day_shifts.append(shift)
                shifts[day] = day_shifts

        self.shifts = shifts

    def get_num_sessions_in_day(self, day: date) -> int:
        shifts = self.p2_store_client.select(
            table='shifts',
            filter={
                'day': day.day,
                'month': day.month,
                'year': day.year,
                'tag': self.tag
            },
            fetch=[
                'shift_id',
                'num_sessions_in_shift'
            ])

        if shifts is None or len(shifts) == 0:
            raise Exception(f"No shifts found for day {day.day}/{day.month}/{day.year}")

        total=0
        for shift in shifts:
            total += shift['num_sessions_in_shift']
        return total

    def get_session_shift(self, duration: int):
        """ looks in all valid days between tomorrow and look_ahead_period
        if there is room for this test...
        """
        today = date.today()
        for delta_day in range(self.config.look_ahead_period):
            day = today + timedelta(days=delta_day+1)
            if self.cal._day_valid(day) is False:
                continue
            if (start_at := self.get_session_shift_in_day(day, duration)) is not None:
                return start_at
        return None

    def get_session_shift_in_day(self, day: date, duration: int):
        """ Given a day and a test duration, checks if there is room
        enough in any of the day test shifts. Checks also that the number
        of sessions already scheduled is less than the max number of
        daily tests configured.

        Returns the shift the test is supposed to run,
        the test should be scheduled to start at the shift start time...
        """

        for shift in self.shifts[day]:
            # see if start+duration is less than end... 
            start_time = datetime(
                year=day.year,
                month=day.month,
                day=day.day,
                hour=shift['start_hour'],
                minute=shift['start_minute']
            )
            end_time = datetime(
                year=day.year,
                month=day.month,
                day=day.day,
                hour=shift['end_hour'],
                minute=shift['end_minute']
            )
            delta_duration = timedelta(minutes=duration)
            new_start = start_time + delta_duration
            if new_start < end_time:
                #okay we found a shift where to run this session...
                total = self.get_num_sessions_in_day(day)
                if (total < self.config.max_tests_per_day):
                    session_start = dict(shift)

                    shift['start_hour'] = new_start.hour
                    shift['start_minute'] = new_start.minute + 1
                    shift['num_sessions_in_shift'] += 1

                    if self.p2_store_client.update_on_conflict(
                        table='shifts',
                        record=shift,
                        conflicting_cols=('day', 'month', 'year', 'shift_id', 'tag')
                        ) == 1:
                        return session_start
        return None
