import os
from datetime import datetime, date

from p2.store.client import Client as StoreClient
from p2.scheduler.Config import Config

class P2test:
    """
    A class to hold all data related with a test session.

    At construction time it fetches the last session and computes the
    expired window if not set. If no previous session is found and expired
    is not set, it is automatically set to 100 days.
    """

    def __init__(self,
                 target:dict,
                 experiment:dict,
                 config:Config,
                 expired:int=None):
        self.p2_store_client = StoreClient(os.environ.get('P2_STORE_URL'))

        self.target = target
        self.experiment = experiment
        self.config = config
        self.rank = 0

        if expired is None:
            self.last_session = self._get_last_session_for_target()
            if self.last_session is not None:
                interval = datetime.today().date() - datetime.fromisoformat(self.last_session['end_time']).date()
                self.expired  = interval.days - self.min_interval()
            else:
                self.expired = 100
        else:
            self.expired = expired


    def _get_last_session_for_target(self):
        """Returns the last session executed for the passed target and experiment"""

        sessions = self.p2_store_client.select(
            table='exp_session',
            filter={
                'filepath': self.experiment['filename'],
                'resource_id': self.target['id']
            })
        if sessions is None or len(sessions) == 0:
            return None

        for s in sessions:
            s['scheduled_day'] = date(s['year'], int(s['month']),
                                      int(s['dom'])).toordinal()

        sessions_sorted = sorted(sessions, key=lambda session:
                                 session['scheduled_day'],
                                 reverse=True)
        for s in sessions_sorted:
            if s['status'] != 'scheduled' and s['status'] != 'deleted':
                return s


    def min_interval(self):
        """computes min interval (in days ) between tests
            of same type on same target """

        target_priority = self._get_target_priority()
        return self.config.target_priority_period[target_priority]


    def _get_target_priority(self):
        """Retrieves the priority index set for this target. If no priority is set,
        returns the lowest index"""

        return self.config.targets_priority.get(self.target['name'],
                            len(self.config.target_priority_period)) - 1
